//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// //
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//
// Author: Ondrej Vondrous
// Department of Telecommunication Engineering
// Czech Technical University in Prague
// Faculty of Electrical Engineering
// 2015
//
//

#include "udg/applications/XMLTrafficGenerator/SocketStorage.h"
#include "inet/transportlayer/contract/udp/UDPControlInfo.h"
#include "inet/transportlayer/contract/tcp/TCPCommand_m.h"

#define MSG "Local:"<<localAddress <<"  TS:"<< simTime() << "  connId()=="<< socket->getConnectionId() << "  "
#define RED "\033[0;31m"
#define iRED "\033[1;31m"
#define GREEN "\033[0;32m"
#define YELLOW "\033[0;33m"
#define BLUE "\033[0;34m"
#define iBLUE "\033[1;34m"
#define PURPLE "\033[0;35m"
#define CYAN "\033[0;36m"
#define DEFAULT "\033[0;00m"

//using namespace inet;

SocketStorage::SocketStorage(int key) {
    int segsize = CONN_ID_LIMIT * sizeof(int);
    /*
     mutex = sem_open("UDGApp_sem1", O_CREAT , 0600, 1);
     if (mutex == SEM_FAILED)
     {
     perror("unable to create semaphore");
     sem_unlink("UDGApp_sem1");
     exit(-1);
     }
     //std::cout << "SEM OPENED\n";
     int * sval;
     sem_getvalue(mutex, sval);
     //std::cout << "sval==" << *sval << endl;
     if (*sval == 0)
     {
     sem_post(mutex);
     }
     */
    shmId = shmget(key, segsize, 0600);
    if (shmId < 0) {
        shmId = shmget(key, segsize, IPC_CREAT | 0600);
        if (shmId < 0) {
            perror("shmget");
            exit(1);
        }
        connIDs = (int *) shmat(shmId, NULL, 0);
        if (connIDs == (int *) -1) {
            perror("shmatt");
            exit(1);
        }
    } else {
        connIDs = (int *) shmat(shmId, NULL, 0);
        if (connIDs == (int *) -1) {
            perror("shmatt");
            exit(1);
        }
    }
    for (int i = 0; i < CONN_ID_LIMIT; i++) {
        connIDs[i] = 0;
    }
}

SocketStorage::~SocketStorage() {
    // TODO Auto-generated destructor stub
    // ToDo napsat destruktor
}

bool SocketStorage::addSocket(inet::TCPSocket *socket, bool initiator) {
    //sem_wait(mutex);
    //std::cout << "LOCK\n";
    if (tcpsockets.count(socket->getConnectionId()) > 0) {
        std::cerr << "Socket with connID: " << RED << socket->getConnectionId()
                << DEFAULT << " is already in SocketStorage!" << endl;
    } else {
        my_tcp_socket_t * my_socket = new my_tcp_socket_t();
        my_socket->active = true;
        my_socket->retired = false;
        my_socket->renew = false;
        my_socket->connPrevState = -1;
        my_socket->connState = -1;
        my_socket->prevState = -1;
        my_socket->last_ts = simTime();
        my_socket->socket = socket;
        my_socket->initiator = initiator;
        if (socket->getRemoteAddress().isUnspecified()) {
            my_socket->listening = true;
            connIDs[socket->getConnectionId()] = 2;
        } else {
            my_socket->listening = false;
            connIDs[socket->getConnectionId()] = 1;
        }
        tcpsockets[socket->getConnectionId()] = my_socket;
        //sem_post(mutex);
        for (my_tcp_it = tcpsockets.begin(); my_tcp_it != tcpsockets.end();
                ++my_tcp_it) {
            inet::TCPSocket * xsocket = my_tcp_it->second->socket;
            if (my_tcp_it->second->renew) {
                if (xsocket->getConnectionId() != socket->getConnectionId()) {
                    if (xsocket->getLocalAddress()
                            == socket->getLocalAddress()) {
                        if (xsocket->getRemoteAddress()
                                == socket->getRemoteAddress()) {
                            if (xsocket->getLocalPort()
                                    == socket->getLocalPort()) {
                                if (xsocket->getRemotePort()
                                        == socket->getRemotePort()) {
                                    std::cout << RED << "RENEW RESET\n";
                                    my_tcp_it->second->renew=false;
                                    my_tcp_it->second->active=false;
                                }
                            }
                        }
                    }
                }
            }
        }

        return true;
    }
    //std::cout << "UNLOCK\n";
    //sem_post(mutex);
    return false;
}

bool SocketStorage::removeSocket(const unsigned int connId) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        delete tcpsockets[connId];
        tcpsockets.erase(connId);
        return true;
    }
    return false;
}

bool SocketStorage::retireSocket(const unsigned int connId) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        if (tcpsockets[connId]->retired != true) {
            tcpsockets[connId]->retired = true;
            tcpsockets[connId]->active = false;
            connIDs[connId] = 0;
            return true;
        } else {
            std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                    << " already retired!" << endl;
        }
    }
    return false;
}

void SocketStorage::setConnState(const unsigned int connId, const int state) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        tcpsockets[connId]->connState = state;
    }
}

void SocketStorage::setListeningSocket(const unsigned int connId,
        const bool state) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        tcpsockets[connId]->listening = state;
    }
}

void SocketStorage::setPrevState(const unsigned int connId, const int state) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        if (tcpsockets[connId]->prevState != state) {
            tcpsockets[connId]->prevState = state;
        }
    }
}

void SocketStorage::setPrevConnState(const unsigned int connId,
        const int state) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        if (tcpsockets[connId]->connPrevState != state) {
            tcpsockets[connId]->connPrevState = state;
        }
    }
}

inet::TCPSocket * SocketStorage::getTCPSocket(const unsigned int connId) {
    if (tcpsockets.count(connId) == 0) {
        //std::cerr << "Socket with connID: " << RED << connId << DEFAULT  << " is not stored in SocketStorage!" << endl;
    } else {
        tcpsockets[connId]->last_ts = simTime();
        tcpsockets[connId]->active = true;
        return tcpsockets[connId]->socket;
    }
    return NULL;
}

inet::TCPSocket * SocketStorage::getTCPSocket(const cMessage * msg) {
    inet::TCPCommand *ind = dynamic_cast<inet::TCPCommand *>(msg->getControlInfo());
    inet::TCPSocket * socket;
    if (!ind) {
        opp_error("No TCPCommand control info in message (not from TCP?)");
    }
    if (tcpsockets.count(ind->getConnId()) == 0) {
        //std::cerr << "Socket with connID: " << RED << ind->getConnId() << DEFAULT << " is not stored in SocketStorage!" << endl;
    } else {
        for (my_tcp_it = tcpsockets.begin(); my_tcp_it != tcpsockets.end();
                ++my_tcp_it) {
            socket = my_tcp_it->second->socket;
            if (!my_tcp_it->second->retired && my_tcp_it->second->active) {
                if (socket->getLocalAddress()
                        == tcpsockets[ind->getConnId()]->socket->getLocalAddress()) {
                    if (socket->getRemoteAddress()
                            == tcpsockets[ind->getConnId()]->socket->getRemoteAddress()) {
                        if (socket->getLocalPort()
                                == tcpsockets[ind->getConnId()]->socket->getLocalPort()) {
                            if (socket->getRemotePort()
                                    == tcpsockets[ind->getConnId()]->socket->getRemotePort()) {
                                my_tcp_it->second->last_ts = simTime();
                                my_tcp_it->second->active = true;
                                return socket;
                            }
                        }
                    }
                }
            }
        }

    }
    return NULL;
}

inet::TCPSocket * SocketStorage::getTCPSocket(const inet::L3Address src_ip,
        const u_int16_t src_port, const inet::L3Address dst_ip,
        const u_int16_t dst_port) {
    inet::TCPSocket * socket;
    for (my_tcp_it = tcpsockets.begin(); my_tcp_it != tcpsockets.end();
            ++my_tcp_it) {
        socket = my_tcp_it->second->socket;
        if (!my_tcp_it->second->retired) {
            if (socket->getLocalAddress() == src_ip) {
                if (socket->getRemoteAddress() == dst_ip) {
                    if (socket->getLocalPort() == src_port) {
                        if (socket->getRemotePort() == dst_port) {
                            my_tcp_it->second->last_ts = simTime();
                            my_tcp_it->second->active = true;
                            return socket;
                        }
                    }
                }
            }
        }
    }
    return NULL;
}

int SocketStorage::getConnState(const unsigned int connId) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        return tcpsockets[connId]->connState;
    }
    return -1;
}

int SocketStorage::getPrevConnState(const unsigned int connId) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        return tcpsockets[connId]->connPrevState;
    }
    return -1;
}

int SocketStorage::getPrevState(const unsigned int connId) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        return tcpsockets[connId]->prevState;
    }
    return -1;
}

bool SocketStorage::listeningSocket(const unsigned int connId) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        return tcpsockets[connId]->listening;
    }
    return false;
}

bool SocketStorage::isEmpty() const {
    return tcpsockets.empty();
}

simtime_t SocketStorage::getLastTS(const unsigned int connId) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        return tcpsockets[connId]->last_ts;
    }
    return 0;

}

bool SocketStorage::connectionCheck() {
    //sem_wait(mutex);
    //std::cout << "LOCK\n";
    unsigned int cnt = 0;
    for (int i = 1; i < CONN_ID_LIMIT; i++) {
        if (connIDs[i] != 0) {
            cnt++;
            if (connIDs[i] == 2) {
                cnt--;
            } else {
                if (tcpsockets.count(i) > 0) {
                    if ((tcpsockets[i]->socket->getState()
                            == inet::TCPSocket::CONNECTED)
                            && ((tcpsockets[i]->last_ts + TIMEOUT) < simTime())) {

                        if ((tcpsockets[i]->initiator)
                                || ((tcpsockets[i]->last_ts + 2*TIMEOUT)
                                        < simTime())) {
                            //std::cout << "Timeout reached - closing socket:" << i << "\tLastTS:" << tcpsockets[i]->last_ts << std::endl;
                            tcpsockets[i]->socket->close();
                            //std::cout << "TIMEOUT >> id == " << i << endl;
                            //connIDs[i] = 0;
                        } else {
                            //std::cout << "Timeout reached - waiting for peer socket to close:" << i << "\tLastTS:" << tcpsockets[i]->last_ts << std::endl;
                        }

                    }
                    if (tcpsockets[i]->socket->getState()
                            == inet::TCPSocket::PEER_CLOSED) {
                        tcpsockets[i]->socket->close();
                        //std::cout << "PEER_CLOSED\n";
                        connIDs[i] = 0;
                    }
                    if ((tcpsockets[i]->socket->getState() == inet::TCPSocket::CLOSED)
                            || (tcpsockets[i]->socket->getState()
                                    == inet::TCPSocket::SOCKERROR)) {
                        connIDs[i] = 0;
                        //cnt--;
                    }
                    if (tcpsockets[i]->socket->getState()
                            == inet::TCPSocket::LISTENING) {
                        connIDs[i] = 2;
                        //cnt--;
                    }
                    if (tcpsockets[i]->socket->getState()
                            == inet::TCPSocket::CONNECTING) {
                        //std::cout << "Connecting" << endl;

                    }
                    if (tcpsockets[i]->renew) {
                        connIDs[i] = 110;
                        cnt++;
                    }
                }
            }
        }
    }

    //sem_post(mutex);
    //std::cout << "Active sockets: " << cnt << std::endl;
    if (cnt == 0) {
        //shmdt((void *) connIDs); //Other workers are still attached, but ...
        //shmctl(shmId, IPC_RMID, NULL);
        return false;
    }
    return true;
}

void SocketStorage::setTimer(const unsigned int connId, simtime_t time) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        tcpsockets[connId]->timer = simTime() + time;
    }
}

void SocketStorage::resetTimer(const unsigned int connId) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        tcpsockets[connId]->timer = 0;
    }
}

simtime_t SocketStorage::getTimer(const unsigned int connId) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        return tcpsockets[connId]->timer;
    }
    return 0;
}

u_int64_t SocketStorage::size() const {
    return tcpsockets.size();
}

bool SocketStorage::alreadyListening(const inet::L3Address src_ip,
        const u_int16_t src_port) {
    inet::TCPSocket * socket;
    for (my_tcp_it = tcpsockets.begin(); my_tcp_it != tcpsockets.end();
            ++my_tcp_it) {
        socket = my_tcp_it->second->socket;
        if (my_tcp_it->second->listening) {
            if (socket->getLocalAddress() == src_ip) {
                if (socket->getLocalPort() == src_port) {
                    return true;
                }
            }
        }
    }
    return false;
}

void SocketStorage::updateLastTS(const unsigned int connId) {
    if (tcpsockets.count(connId) == 0) {
        std::cerr << "Socket with connID: " << RED << connId << DEFAULT
                << " is not stored in SocketStorage!" << endl;
    } else {
        tcpsockets[connId]->last_ts = simTime();
    }
}

void SocketStorage::renewSockets() {
    inet::TCPSocket * socket;
    for (my_tcp_it = tcpsockets.begin(); my_tcp_it != tcpsockets.end();
            ++my_tcp_it) {
        socket = my_tcp_it->second->socket;

        if (my_tcp_it->second->renew) {
            my_tcp_it->second->renew = false;
            inet::L3Address src = socket->getLocalAddress();
            inet::L3Address dest = socket->getRemoteAddress();
            int src_port = socket->getLocalPort();
            int dst_port = socket->getRemotePort();
            socket->renewSocket();
            socket->bind(src, src_port);
            socket->connect(dest, dst_port);
            EV << "issuing OPEN command\n";
            if (addSocket(socket, true)) {
                setPrevState(socket->getConnectionId(), inet::TCPSocket::CONNECTING);
            } else {
                std::cerr << "AddSocket Collision!" << endl;
            }
        }
    }
}

void SocketStorage::renewSocket(const unsigned int connId) {
    tcpsockets[connId]->renew = true;
}

bool SocketStorage::renewSocket(cMessage * msg) {
    inet::TCPCommand *ind = dynamic_cast<inet::TCPCommand *>(msg->getControlInfo());
    if (!ind) {
        opp_error("No TCPCommand control info in message (not from TCP?)");
    }
    if (tcpsockets.count(ind->getConnId()) == 0) {
        std::cerr << "Socket with connID: " << RED << ind->getConnId()
                << DEFAULT << " is not stored in SocketStorage!" << endl;
    } else {
        tcpsockets[ind->getConnId()]->renew = true;
        return tcpsockets[ind->getConnId()]->initiator;
    }
    return false;
}

