//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// //
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//
// Author: Ondrej Vondrous
// Department of Telecommunication Engineering
// Czech Technical University in Prague
// Faculty of Electrical Engineering
// 2015
//
//

#include <fcntl.h>
#include <map>
#include "inet/networklayer/common/L3Address.h"
#include <sys/types.h>
#include "inet/transportlayer/contract/tcp/TCPSocket.h"
#include "inet/transportlayer/contract/udp/UDPSocket.h"
#include <vector>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <semaphore.h>



#ifndef SOCKETSTORAGE_H_
#define SOCKETSTORAGE_H_

#define TIMEOUT 1800
#define RECOVERY_TIMEOUT 240
#define CONN_ID_LIMIT 100000 //400 kiB


struct my_tcp_socket_t{
    bool active;
    bool listening;
    bool retired;
    bool initiator;
    bool renew;
    int prevState;
    int connState;
    int connPrevState;
    simtime_t timer;
    simtime_t last_ts;
    inet::TCPSocket * socket;
};

struct my_udp_socket_t{
    bool active;
    bool listening;
    bool retired;
    int prevState;
    inet::UDPSocket * socket;
};

class SocketStorage {
public:
    SocketStorage(int key);
    virtual ~SocketStorage();

    bool addSocket(inet::TCPSocket * socket,bool initiator);
    bool addSocket(inet::UDPSocket * socket);
    bool removeSocket(const unsigned int connId);
    bool retireSocket(const unsigned int connId);
    void setPrevState(const unsigned int connId, const int state);
    void setConnState(const unsigned int connId, const int state);
    void setPrevConnState(const unsigned int connId, const int state);
    void setListeningSocket(const unsigned int connId, const bool state);
    void resetTimer(const unsigned int connId);
    void renewSockets(void);
    bool renewSocket(cMessage * msg);
    void renewSocket(const unsigned int connId);
    void setTimer(const unsigned int connId,simtime_t time);
    inet::TCPSocket * getTCPSocket(const cMessage * msg);
    inet::TCPSocket * getTCPSocket(const unsigned int conId);
    inet::TCPSocket * getTCPSocket(const inet::L3Address src_ip,const u_int16_t src_port,const inet::L3Address dst_ip, const u_int16_t dst_port);
    inet::UDPSocket * getUDPSocket(cMessage * msg) const;
    inet::UDPSocket * getUDPSocket(const inet::L3Address src_ip,const u_int16_t src_port,const inet::L3Address dst_ip, const u_int16_t dst_port) const;
    int getPrevState(const unsigned int connId);
    int getConnState(const unsigned int connId);
    int getPrevConnState(const unsigned int connId);
    simtime_t getTimer(const unsigned int connId);
    simtime_t getLastTS(const unsigned int connId);
    void updateLastTS(const unsigned int connId);
    bool listeningSocket(const unsigned int connId);
    bool isEmpty(void) const;
    u_int64_t size(void) const;
    bool connectionCheck(void);
    bool alreadyListening(const inet::L3Address src_ip,const u_int16_t src_port);
    //void socketGC(void);
    //void print(void);

protected:
    std::map<unsigned int, my_tcp_socket_t *> tcpsockets; //[connId] --> MyTCPSSocket
    std::map<unsigned int, my_udp_socket_t *> udpsockets; //[connId] --> MyUDPSSocket
private:
    std::map<unsigned int, my_tcp_socket_t *>::iterator my_tcp_it;
    int shmflgs;
    int shmId;
    sem_t * mutex;
    int retval;
    int * connIDs;
};


#endif /* SOCKETSTORAGE_H_ */
