//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// //
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//
// Author: Ondrej Vondrous
// Department of Telecommunication Engineering
// Czech Technical University in Prague
// Faculty of Electrical Engineering
// 2015
//
//

#include "udg/applications/XMLTrafficGenerator/TCPUDGApp.h"
#include "inet/networklayer/common/L3AddressResolver.h"
#include "inet/networklayer/common/L3Address.h"
#include "inet/common/ModuleAccess.h"
#include "inet/transportlayer/contract/tcp/TCPSocket.h"
#include "inet/transportlayer/tcp/TCPConnection.h"
#include "inet/transportlayer/contract/tcp/TCPCommand_m.h"

Define_Module(TCPUDGApp);

//TODO CHECK MGS_KINDs IDs (for collisions)
#define MSGKIND_CONNECT  101            //redefined may be in collision
#define MSGKIND_SEND     102            //redefined may be in collision
#define MSGKIND_CLOSE    103              //redefined may be in collision
#define MSGKIND_CONN_CHECK    104         //My MSG
#define MSGKIND_END    109               //My MSG
#define MSGKIND_RENEW_SOCKET    110               //My MSG

//using namespace inet;

TCPUDGApp::TCPUDGApp() {
    timeoutMsg = NULL;
    connTimeoutMsg = NULL;
}

TCPUDGApp::~TCPUDGApp() {
    cancelAndDelete(timeoutMsg);
    cancelAndDelete(connTimeoutMsg);

}

void TCPUDGApp::initialize(int stage) {
    TCPUDGAppBase::initialize(stage);
    eedSignal = registerSignal("endToEnd");
    if (stage == inet::INITSTAGE_LOCAL) {
        //if (stage == 0) {
        activeOpen = par("active");
        tOpen = par("tOpen");
        if (tOpen < 0)
            tOpen = 0;   //when some sort of distribution function is used, ....
        tSend = par("tSend");
        if (tSend < 0)
            tSend = 0;
        tClose = par("tClose");
        if (tClose < 0)
            tClose = 0;
    }
    if (stage == inet::INITSTAGE_APPLICATION_LAYER) {
        //if (stage == 3) {
        //Net is initialized at stage 2
        localAddress = inet::L3AddressResolver().addressOf(
                this->getParentModule());
        udg = new UDGConfig(par("udgconfig"), localAddress, tcp_t);
        udg_profiles = udg->getUDGProfiles(localAddress, tcp_t);
        for (unsigned int i = 0; i < udg_profiles.size(); i++) {
            udg_profiles[i]->stime = tOpen.dbl() + tSend.dbl();
        }

        for (unsigned int i = 0; i < udg_profiles.size(); i++) {
            if (udg_profiles[i]->dst_ip == localAddress) {
                if (!sockets->alreadyListening(localAddress,
                        udg_profiles[i]->dst_port)) {
                    inet::TCPSocket * socket = new inet::TCPSocket();
                    socket->setOutputGate(gate("tcpOut"));
                    socket->readDataTransferModePar(*this);
                    socket->setCallbackObject(this);
                    socket->bind(localAddress, udg_profiles[i]->dst_port);
                    socket->listen();
                    if (sockets->addSocket(socket, false)) {
                        sockets->setPrevState(socket->getConnectionId(),
                                inet::TCPSocket::LISTENING);
                        sockets->setListeningSocket(socket->getConnectionId(),
                                true);
                        cout << "Host listening at: "
                                << socket->getLocalAddress() << ":"
                                << socket->getLocalPort() << " is ready."
                                << endl;
                    } else {
                        //
                    }
                }
            }
        }
        setStatusString("listening");
        timeoutMsg = new cMessage("timer");
        connTimeoutMsg = new cMessage("connTimer");

        timeoutMsg->setKind(MSGKIND_CONNECT);
        scheduleAt(tOpen, timeoutMsg);
        connTimeoutMsg->setKind(MSGKIND_CONN_CHECK);
        scheduleAt(tOpen + TIMEOUT, connTimeoutMsg);

    }
}

//Only initial connect
void TCPUDGApp::connect() {
    for (unsigned int i = 0; i < udg_profiles.size(); i++) {
        if (inet::L3Address(udg_profiles[i]->src_ip) == localAddress) {
            inet::TCPSocket * socket = new inet::TCPSocket();
            socket->setOutputGate(gate("tcpOut"));
            socket->setCallbackObject(this);
            socket->readDataTransferModePar(*this);
            socket->bind(localAddress, udg_profiles[i]->src_port);
            socket->connect(udg_profiles[i]->dst_ip, udg_profiles[i]->dst_port);
            EV << "issuing OPEN command\n";
            setStatusString("connecting");
            numSessions++;
            if (sockets->addSocket(socket, true)) {
                sockets->setPrevState(socket->getConnectionId(),
                        inet::TCPSocket::CONNECTING);
                emit(connectSignal, 1L);
                if (!sockets->listeningSocket(socket->getConnectionId())) {
                    socket = new inet::TCPSocket();
                    socket->setOutputGate(gate("tcpOut"));
                    socket->readDataTransferModePar(*this);
                    socket->setCallbackObject(this);
                    socket->bind(localAddress, udg_profiles[i]->src_port);
                    socket->listen();
                    if (sockets->addSocket(socket, false)) {
                        sockets->setPrevState(socket->getConnectionId(),
                                inet::TCPSocket::LISTENING);
                        sockets->setListeningSocket(socket->getConnectionId(),
                                true);
                    }
                }
            } else {
                cerr << RED << "AddSocket Collision!" << DEFAULT << endl;
            }
        }
    }
}

void TCPUDGApp::handleMessage(cMessage *msg) {

    inet::TCPCommand * cmd;
    //cMessage * xmsg;
    inet::L3Address local, remote;
    //bool process = true;

    //int srcport, dstport;
    if (msg->isSelfMessage()) {
        if (msg->isPacket()) {
            if (msg->getKind() == inet::TCP_C_SEND) {
                cmd = check_and_cast<inet::TCPCommand *>(msg->getControlInfo());
                inet::TCPSocket *socket;
                bytesSent += ((cPacket *) msg)->getByteLength();
                emit(sentPkSignal, (cPacket *) msg);
                socket = sockets->getTCPSocket(msg);
                if (socket != NULL) {
                    if (socket->getState() == inet::TCPSocket::CONNECTED) {
                        msg->removeControlInfo();
                        Udg_packet *pkt = check_and_cast<Udg_packet *>(msg);
                        pkt->setSent_time(simTime());
                        sockets->updateLastTS(socket->getConnectionId());
                        socket->send(msg);
                        sockets->setPrevState(socket->getConnectionId(),
                                inet::TCPSocket::CONNECTED);
                        return;
                    } else {

                        if (socket->getState() == inet::TCPSocket::CONNECTING) {
                            scheduleAt(simTime() + 1, msg);
                            sockets->setPrevState(socket->getConnectionId(),
                                    socket->getState());
                            return;
                        }
                    }
                    sockets->setPrevState(socket->getConnectionId(),
                            socket->getState());
                } else {
                    if (!timeoutMsg->isScheduled()) {
                        timeoutMsg->setKind(MSGKIND_RENEW_SOCKET);
                        if (sockets->renewSocket(msg)) {
                            scheduleAt(simTime() + RECOVERY_TIMEOUT,
                                    timeoutMsg);
                        } else {
                            scheduleAt(simTime() + RECOVERY_TIMEOUT + 30,
                                    timeoutMsg);
                        }
                    }
                }
                scheduleAt(simTime() + 1, msg);
            }
        } else {
            handleTimer(msg);
        }
    } else {
        inet::TCPSocket * socket = sockets->getTCPSocket(msg);
        if (!socket) {
            socket = new inet::TCPSocket(msg);
            socket->setCallbackObject(this);
            socket->readDataTransferModePar(*this);
            socket->setOutputGate(gate("tcpOut"));
            if (socket->getRemoteAddress().isUnspecified()) {
                cerr << "Unspecified address for message: " << msg << "\t"
                        << msg->getKind() << endl;
            } else {
                if (!sockets->addSocket(socket, false)) {
                    throw cRuntimeError("Socket == NULL, but already exists");
                }
            }
        }
        inet::TCPCommand *ind =
                dynamic_cast<inet::TCPCommand *>(msg->getControlInfo());
        if (!ind) {
            opp_error("No TCPCommand control info in message (not from TCP?)");
        }
        socket->processMessage(msg);
    }

}

void TCPUDGApp::handleTimer(cMessage *msg) {
    ev << msg->getKind() << "\t" << msg << endl;
    switch (msg->getKind()) {
    case MSGKIND_CONNECT:
        EV << "opening connection\n";
        if (activeOpen)
            connect(); // sending will be scheduled from socketEstablished()
        else
            ; //TODO
        break;

    case MSGKIND_SEND:
        if (sockets->size()) {
            sendData();
        }
        break;

    case MSGKIND_RENEW_SOCKET:
        if (sockets->size()) {
            sockets->renewSockets();
        }
        break;

    case MSGKIND_CONN_CHECK:
        if (!sockets->connectionCheck()) {
            connTimeoutMsg->setKind(MSGKIND_END);
            scheduleAt(simTime() + 10, msg); //10sec grace period for last TCP connection 4way handshake.
        } else {
            scheduleAt(simTime() + 1, msg);
        }
        break;
    case MSGKIND_END:
        endSimulation();
        break;

    case MSGKIND_CLOSE:
        //Ignore message - other side can still comunicate, and there can be also packets to send in reaction ....

        break;

    default:
        throw cRuntimeError("Invalid timer msg: kind=%d", msg->getKind());
    }
}

void TCPUDGApp::sendData() {
    double tSend;
    unsigned int index;
    unsigned int count;
    unsigned int pkt_id;
    index = 0;
    tSend = 0;
    count = 0;
    inet::L3Address local, remote;
    int srcport, dstport;
    //inet::TCPCommand * cmd;
    //cMessage * xmsg;

    //find packet to send
    for (unsigned int i = 0; i < udg_profiles.size(); i++) {
        if (inet::L3Address(udg_profiles[i]->src_ip) == localAddress) {
            if (!udg_profiles[i]->sd_send_data.empty()) {
                count += udg_profiles[i]->sd_send_data.size();
                if ((tSend == 0)
                        || ((udg_profiles[i]->sd_send_data.front()->time
                                + udg_profiles[i]->stime) < tSend)) {
                    tmp_packet = udg_profiles[i]->sd_send_data.front();
                    tSend = tmp_packet->time + udg_profiles[i]->stime;
                    pkt_id = tmp_packet->id;
                    index = i;
                }
            }
        } else {
            if (!udg_profiles[i]->ds_send_data.empty()) {
                count += udg_profiles[i]->ds_send_data.size();
                if ((tSend == 0)
                        || ((udg_profiles[i]->ds_send_data.front()->time
                                + udg_profiles[i]->stime) < tSend)) {
                    tmp_packet = udg_profiles[i]->ds_send_data.front();
                    tSend = tmp_packet->time + udg_profiles[i]->stime;
                    pkt_id = tmp_packet->id;
                    index = i;
                }
            }
        }
    }
//send packet
    if (tSend) {
        inet::TCPSocket * socket = sockets->getTCPSocket(
                udg_profiles[index]->src_ip, udg_profiles[index]->src_port,
                udg_profiles[index]->dst_ip, udg_profiles[index]->dst_port);
        if (socket == NULL) {
            if (udg_profiles[index]->src_ip == localAddress) {
                local = udg_profiles[index]->src_ip;
                remote = udg_profiles[index]->dst_ip;
                srcport = udg_profiles[index]->src_port;
                dstport = udg_profiles[index]->dst_port;
            } else {
                local = udg_profiles[index]->dst_ip;
                remote = udg_profiles[index]->src_ip;
                srcport = udg_profiles[index]->dst_port;
                dstport = udg_profiles[index]->src_port;
            }
            if (!timeoutMsg->isScheduled()) {
                cancelEvent(timeoutMsg);
            }
            timeoutMsg->setKind(MSGKIND_SEND);
            scheduleAt(simTime() + 1, timeoutMsg);

        } else {

            if (socket->getState() == inet::TCPSocket::CONNECTED) {
                if (inet::L3Address(udg_profiles[index]->src_ip)
                        == localAddress) {
                    udg_profiles[index]->sd_sent_xdata.push(tmp_packet);
                    udg_profiles[index]->sd_send_data.pop();
                    scheduleAt(simTime() + 0.00000001,
                            createDataPacket(socket, pkt_id,
                                    udg_profiles[index]->id, tmp_packet->size));
                } else {
                    udg_profiles[index]->ds_sent_xdata.push(tmp_packet);
                    udg_profiles[index]->ds_send_data.pop();
                    scheduleAt(simTime() + 0.0000001,
                            createDataPacket(socket, pkt_id,
                                    udg_profiles[index]->id, tmp_packet->size));
                }
            } else {
                if (!timeoutMsg->isScheduled()) {
                    timeoutMsg->setKind(MSGKIND_SEND);
                    scheduleAt(simTime() + 1, timeoutMsg);
                }
            }
            sockets->setPrevState(socket->getConnectionId(),
                    socket->getState());
        }
    }
//set Timeout
    ev << "cout:"<<count<<endl;
    if (count > 1) {
        tSend = 0;
        for (unsigned int i = 0; i < udg_profiles.size(); i++) {
            if (inet::L3Address(udg_profiles[i]->src_ip) == localAddress) {
                if (!udg_profiles[i]->sd_send_data.empty()) {
                    if ((tSend == 0)
                            || (udg_profiles[i]->sd_send_data.front()->time
                                    + udg_profiles[i]->stime < tSend)) {
                        tmp_packet = udg_profiles[i]->sd_send_data.front();
                        tSend = tmp_packet->time + udg_profiles[i]->stime;
                        udg_profiles[i]->sd_sent_xdata.push(udg_profiles[i]->sd_send_data.front());
                        udg_profiles[i]->sd_send_data.pop();

                    }
                }
            } else {
                if (!udg_profiles[i]->ds_send_data.empty()) {
                    if ((tSend == 0)
                            || (udg_profiles[i]->ds_send_data.front()->time
                                    + udg_profiles[i]->stime < tSend)) {
                        tmp_packet = udg_profiles[i]->ds_send_data.front();
                        tSend = tmp_packet->time + udg_profiles[i]->stime;
                        udg_profiles[i]->ds_sent_xdata.push(udg_profiles[i]->ds_send_data.front());
                        udg_profiles[i]->ds_send_data.pop();

                    }
                }
            }
        }
        if (!timeoutMsg->isScheduled()) {
            EV << "sheduling next send to: " << tSend << endl;
            scheduleAt(std::max(SimTime(tSend), simTime()), timeoutMsg);
        }
    } else {
//no mora data to send -> close // todo -> reply data can arrive!
        if (timeoutMsg->isScheduled()) {
            cancelEvent(timeoutMsg);
        }
        //cout << RED << "SUPER close" << DEFAULT << endl;
        timeoutMsg->setKind(MSGKIND_CLOSE);
        scheduleAt(std::max(tClose, simTime()), timeoutMsg);
    }
}

cPacket *TCPUDGApp::createDataPacket(const inet::TCPSocket * socket,
        const unsigned int pkt_id, const unsigned int profile_id,
        const long sendBytes) {
    inet::TCPCommand *cmd;
    int connId = socket->getConnectionId();
    switch (socket->getDataTransferMode()) {
    case inet::TCP_TRANSFER_OBJECT: {
        Udg_packet *msg = static_cast<Udg_packet *>(newUdgPacket());
        msg->setProfile_id(profile_id);
        msg->setPacket_id(pkt_id);
        msg->setByteLength(sendBytes);
        cmd = (inet::TCPCommand *) msg->getControlInfo();
        if (cmd == NULL) {
            cmd = new inet::TCPCommand();
        }
        cmd->setConnId(connId);
        msg->setControlInfo(cmd);
        msg->setKind(inet::TCP_C_SEND);
        return msg;
    }

    case inet::TCP_TRANSFER_BYTECOUNT: {
        throw cRuntimeError(
                "Fatal error occurred: TCP_TRANSFER_BYTECOUNT mode is not supported by TCPUDGApp");
    }

    case inet::TCP_TRANSFER_BYTESTREAM: {
        throw cRuntimeError(
                "Fatal error occurred: TCP_TRANSFER_BYTECOUNT mode is not supported by TCPUDGApp");
    }

    default:
        throw cRuntimeError("Invalid TCP data transfer mode: %d",
                socket->getDataTransferMode());
    }
}

void TCPUDGApp::socketEstablished(int connId, void *ptr) {
    inet::TCPSocket * socket;
    EV << "connected: " << connId << "\n";
    setStatusString("connected");
    double sd_tSend, ds_tSend;
    ds_tSend = -1;
    sd_tSend = -1;
    socket = sockets->getTCPSocket(connId);
    renewSendData(connId);
    if (socket != NULL) {
        for (unsigned int i = 0; i < udg_profiles.size(); i++) {
            if (inet::L3Address(udg_profiles[i]->src_ip) == localAddress) {
                if (udg_profiles[i]->src_port == socket->getLocalPort()) {
                    if (!udg_profiles[i]->sd_send_data.empty()) {
                        if ((sd_tSend < 0)
                                || (udg_profiles[i]->sd_send_data.front()->time
                                        + udg_profiles[i]->stime < sd_tSend)) {
                            tmp_packet = udg_profiles[i]->sd_send_data.front();
                            sd_tSend = tmp_packet->time
                                    + udg_profiles[i]->stime;
                        }
                    }

                }
            } else {
                if (udg_profiles[i]->dst_port == socket->getLocalPort()) {
                    if (!udg_profiles[i]->ds_send_data.empty()) {
                        if ((ds_tSend < 0)
                                || (udg_profiles[i]->ds_send_data.front()->time
                                        + udg_profiles[i]->stime < ds_tSend)) {
                            tmp_packet = udg_profiles[i]->ds_send_data.front();
                            ds_tSend = tmp_packet->time
                                    + udg_profiles[i]->stime;
                        }
                    }
                }
            }
        }
        if (!timeoutMsg->isScheduled()) {
            if (sd_tSend > -1) {
                timeoutMsg->setKind(MSGKIND_SEND);
                scheduleAt(std::max(SimTime(sd_tSend), simTime()), timeoutMsg);
            }
            if (ds_tSend > -1) {
                timeoutMsg->setKind(MSGKIND_SEND);
                scheduleAt(std::max(SimTime(ds_tSend), simTime()), timeoutMsg);
            }
        }
    } else {
        throw cRuntimeError("NULL socket in ESTABLISHED Section!");
    }
}

void TCPUDGApp::socketDataArrived(int connId, void *ptr, cPacket *msg,
        bool urgent) {
    packetsRcvd++;
    bytesRcvd += msg->getByteLength();
    tmp_packet = NULL;
    inet::TCPSocket * socket = sockets->getTCPSocket(msg);
    if (socket != NULL) {
        sockets->updateLastTS(socket->getConnectionId());
    }
    Udg_packet *reply_pkt;
    Udg_packet *pkt = check_and_cast<Udg_packet *>(msg);
    /*
     if (socket!=NULL){
     cout << socket->getLocalAddress()<<":"<<socket->getLocalPort()<<"\t"<<pkt->getProfile_id()<<"/"<<pkt->getPacket_id()<<endl;
     }
     if (pkt->getProfile_id()==0 and pkt->getPacket_id()==0){
     cout <<"aaa"<<endl;
     }
     */
    ev << "Socket data arrived" << msg << endl;
    emit(rcvdPkSignal, pkt);
    emit(eedSignal, pkt->getArrivalTime() - pkt->getSent_time());
    //std::cout << pkt->getArrivalTime() - pkt->getSent_time() <<endl;
    bytesRcvd += pkt->getByteLength();
    for (unsigned int i = 0; i < udg_profiles.size(); i++) {
        if (udg_profiles[i]->reply_data.count(pkt->getProfile_id()) > 0) {
            tmp_pkt_map = udg_profiles[i]->reply_data[pkt->getProfile_id()];
            if (tmp_pkt_map->count(pkt->getPacket_id()) > 0) {
                while (!(*tmp_pkt_map)[pkt->getPacket_id()]->empty()) {
                    tmp_packet = (*tmp_pkt_map)[pkt->getPacket_id()]->front();
                    //check direction SD/DS of reply packet

                    if (udg_profiles[i]->src_ip == localAddress) {
                        if (tmp_packet->dir == ds) {
                            cout << "ds! SRC IP: " << localAddress << "\tid: "
                                    << tmp_packet->id << "\t Profile: "
                                    << udg_profiles[i]->id << endl;
                            throw cRuntimeError(
                                    "Fatal error occurred: Wrong packet direction in UDG XML");
                        }
                        socket = sockets->getTCPSocket(udg_profiles[i]->src_ip,
                                udg_profiles[i]->src_port,
                                udg_profiles[i]->dst_ip,
                                udg_profiles[i]->dst_port);
                    } else {
                        if (tmp_packet->dir == sd) {
                            cout << "sd! DST IP: " << localAddress << "\tid: "
                                    << tmp_packet->id << endl;
                            throw cRuntimeError(
                                    "Fatal error occurred: Wrong packet direction in UDG XML");
                        }
                        socket = sockets->getTCPSocket(udg_profiles[i]->dst_ip,
                                udg_profiles[i]->dst_port,
                                udg_profiles[i]->src_ip,
                                udg_profiles[i]->src_port);
                    }
                    if (socket == NULL) {
                        throw cRuntimeError(
                                "Fatal error occured in socketDataArrived(): NULL SENDING SOCKET!");
                    }
                    (*tmp_pkt_map)[pkt->getPacket_id()]->pop();
                    reply_pkt = new Udg_packet("UDG Reply");
                    reply_pkt->setPacket_id(tmp_packet->id);
                    reply_pkt->setProfile_id(udg_profiles[i]->id);
                    reply_pkt->setByteLength(tmp_packet->size);
                    reply_pkt->setKind(inet::TCP_C_SEND);
                    inet::TCPSendCommand *cmd = new inet::TCPSendCommand();
                    cmd->setConnId(socket->getConnectionId());
                    reply_pkt->setControlInfo(cmd);
                    // Always add ONE u_sec as processing delay. (HACK - simullation sometimes fails if cumulative delay is 0);
                    scheduleAt(simTime() + tmp_packet->time + 0.000001,
                            reply_pkt);
                }
            }
        }
    }
    delete msg;
}

void TCPUDGApp::socketClosed(int connId, void *ptr) {
    TCPUDGAppBase::socketClosed(connId, ptr);
    sockets->retireSocket(connId);
    if (sockets->getPrevState(connId) == inet::TCPSocket::SOCKERROR) {
        sockets->renewSocket(connId);
        if (timeoutMsg->isScheduled()) {
            cancelEvent(timeoutMsg);
        }
        timeoutMsg->setKind(MSGKIND_RENEW_SOCKET);
        scheduleAt(simTime() + RECOVERY_TIMEOUT, timeoutMsg);
    }
}

void TCPUDGApp::renewSendData(const int connId) {
    inet::TCPSocket * socket = sockets->getTCPSocket(connId);
    if (socket != NULL) {
        for (unsigned int i = 0; i < udg_profiles.size(); i++) {
            if ((inet::L3Address(udg_profiles[i]->src_ip) == localAddress)
                    && (udg_profiles[i]->src_port == socket->getLocalPort())
                    && (inet::L3Address(udg_profiles[i]->dst_ip)
                            == socket->getRemoteAddress())
                    && (udg_profiles[i]->dst_port == socket->getRemotePort())) {
                while (!udg_profiles[i]->sd_send_data.empty()) {
                    udg_profiles[i]->sd_sent_xdata.push(
                            udg_profiles[i]->sd_send_data.front());
                    udg_profiles[i]->sd_send_data.pop();
                }
                while (!udg_profiles[i]->ds_send_data.empty()) {
                    udg_profiles[i]->ds_sent_xdata.push(
                            udg_profiles[i]->ds_send_data.front());
                    udg_profiles[i]->ds_send_data.pop();
                }
                while (!udg_profiles[i]->sd_sent_xdata.empty()) {
                    udg_profiles[i]->sd_send_data.push(
                            udg_profiles[i]->sd_sent_xdata.front());
                    udg_profiles[i]->sd_sent_xdata.pop();
                }
                while (!udg_profiles[i]->ds_sent_xdata.empty()) {
                    udg_profiles[i]->ds_send_data.push(
                            udg_profiles[i]->ds_sent_xdata.front());
                    udg_profiles[i]->ds_sent_xdata.pop();
                }
                udg_profiles[i]->stime = simTime().dbl();
            }
        }

    } else {
        cerr << RED << "SOCKET not found" << DEFAULT << endl;
    }
}

void TCPUDGApp::socketFailure(int connId, void *ptr, int code) {
    TCPUDGAppBase::socketFailure(connId, ptr, code);
    inet::TCPSocket * socket = sockets->getTCPSocket(connId);
    inet::TCPCommand * cmd;
    cMessage * xmsg;
    switch (code) {
    case inet::TCP_I_CONNECTION_REFUSED:
        cerr << localAddress << "\t" << RED << "CONN REFUSED" << "\t" << connId
                << DEFAULT << endl;
        break;
    case inet::TCP_I_CONNECTION_RESET:
        cerr << localAddress << "\t" << RED << "CONN RESET" << "\t" << connId
                << DEFAULT << endl;
        break;
    case inet::TCP_I_TIMED_OUT:
        cerr << localAddress << "\t" << RED << "CONN TIMEOUT" << "\t" << connId
                << DEFAULT << endl;
        break;
    default:
        cerr << RED << "CONN ???????" << "\t" << connId << DEFAULT << endl;
        break;
    }
    sockets->retireSocket(connId);
    sockets->setPrevState(connId, inet::TCPSocket::SOCKERROR);
}

void TCPUDGApp::finish() {
    EV << getFullPath() << ": received " << bytesRcvd << " bytes in "
              << packetsRcvd << " packets\n";
    recordScalar("bytesRcvd", bytesRcvd);
    recordScalar("bytesSent", bytesSent);
}

Udg_packet * TCPUDGApp::newUdgPacket() {
    return new Udg_packet("UDG_Data_Packet");
}
