//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// //
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//
// Author: Ondrej Vondrous
// Department of Telecommunication Engineering
// Czech Technical University in Prague
// Faculty of Electrical Engineering
// 2015
//
//


#ifndef TCPUDGAPP_H_
#define TCPUDGAPP_H_

#include "INETDefs.h"
#include "inet/common/lifecycle/LifecycleOperation.h"
#include "udg/applications/XMLTrafficGenerator/TCPUDGAppBase.h"
#include "inet/common/lifecycle/NodeOperations.h"
#include "inet/common/lifecycle/NodeStatus.h"
#include <map>


class TCPUDGApp: public TCPUDGAppBase {
public:
    TCPUDGApp();
    virtual ~TCPUDGApp();
protected:
    bool activeOpen;
    simtime_t tOpen;
    simtime_t tSend;
    simtime_t tClose;
    cMessage *timeoutMsg;
    cMessage *connTimeoutMsg;
    //NodeStatus *nodeStatus;
protected:
    inet::TCPStatusInfo * status; //indicates last status
    int statusConnId; //for storing imediate connection status state.
    simsignal_t eedSignal;

    virtual void connect();
    virtual void handleMessage(cMessage *msg);
/*
    virtual bool isNodeUp();
    virtual bool handleOperationStage(LifecycleOperation *operation, int stage, IDoneCallback *doneCallback);
*/
    virtual int numInitStages() const {
        return 16; //NUM_INIT_STAGES for Inet 3.x
    }
    virtual void initialize(int stage);
    virtual void finish();
    //virtual cPacket *createDataPacket(long sendBytes);
    virtual cPacket *createDataPacket(const inet::TCPSocket * socket,const unsigned int pkt_id,const unsigned int profile_id, const long sendBytes);
    virtual void sendData();
    virtual void handleTimer(cMessage *msg);
    virtual void socketEstablished(int connId, void *yourPtr);
    virtual void socketDataArrived(int connId, void *yourPtr, cPacket *msg,
            bool urgent);
    virtual void socketClosed(int connId, void *yourPtr);
    virtual void socketFailure(int connId, void *yourPtr, int code);
    virtual Udg_packet * newUdgPacket();
    virtual void renewSendData(const int connId);
};
#endif /* TCPUDGAPP_H_ */



