//
// Copyright 2004 Andras Varga
//
// This library is free software, you can redistribute it and/or modify
// it under  the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation;
// either version 2 of the License, or any later version.
// The library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details.
//
// Modified by: Ondrej Vondrous
// Department of Telecommunication Engineering
// Czech Technical University in Prague
// Faculty of Electrical Engineering
// 2015
//

#include "udg/applications/XMLTrafficGenerator/TCPUDGAppBase.h"


simsignal_t TCPUDGAppBase::connectSignal = registerSignal("connect");
simsignal_t TCPUDGAppBase::rcvdPkSignal = registerSignal("rcvdPk");
simsignal_t TCPUDGAppBase::sentPkSignal = registerSignal("sentPk");

void TCPUDGAppBase::initialize(int stage) {
    cSimpleModule::initialize(stage);

    if (stage == 0) {
        numSessions = numBroken = packetsSent = packetsRcvd = bytesSent =
                bytesRcvd = 0;

        WATCH(numSessions);
        WATCH(numBroken);
        WATCH(packetsSent);
        WATCH(packetsRcvd);
        WATCH(bytesSent);
        WATCH(bytesRcvd);
        sockets=new SocketStorage(par("key"));
    }
}


void TCPUDGAppBase::handleMessage(cMessage *msg) {
    if (msg->isSelfMessage())
        handleTimer(msg);
    else {
        /*
         TCPSocket *socket = socketMap->findSocketFor(msg);
         if (!socket) {
         // new connection -- create new socket object and server process
         socket = new TCPSocket(msg);
         socket->setOutputGate(gate("tcpOut"));
         socket->setCallbackObject(this);
         socketMap.addSocket(socket);
         }
         socket->processMessage(msg);
         */
    }

}

void TCPUDGAppBase::connect() {
}

void TCPUDGAppBase::close(inet::TCPSocket * socket) {
    setStatusString("closing");
    EV << "issuing CLOSE command\n";
    socket->close();
    emit(connectSignal, -1L);
}

void TCPUDGAppBase::sendPacket(inet::TCPSocket * socket, cPacket *msg) {
    int numBytes = msg->getByteLength();
    emit(sentPkSignal, msg);
    socket->send(msg);
    //cout << "SP .::. socket.getConnectionId() "<<   socket->getConnectionId()<<endl;
    packetsSent++;
    bytesSent += numBytes;
}

void TCPUDGAppBase::setStatusString(const char *s) {
    if (ev.isGUI())
        getDisplayString().setTagArg("t", 0, s);
}

void TCPUDGAppBase::socketEstablished(int, void *) {
// *redefine* to perform or schedule first sending
    EV << "connected\n";
    setStatusString("connected");
}

void TCPUDGAppBase::socketDataArrived(int, void *, cPacket *msg, bool) {
// *redefine* to perform or schedule next sending
    packetsRcvd++;
    bytesRcvd += msg->getByteLength();
    emit(rcvdPkSignal, msg);
    delete msg;
}

void TCPUDGAppBase::socketPeerClosed(inet::TCPSocket * socket, int, void *) {
// close the connection (if not already closed)
    if (socket->getState() == inet::TCPSocket::PEER_CLOSED) {
        EV << "remote TCP closed, closing here as well\n";
        //cerr << "remote TCP closed, closing here as well\n";
        close(socket);
    }
}

void TCPUDGAppBase::socketClosed(int, void *) {
// *redefine* to start another session etc.
    EV << "connection closed\n";
    setStatusString("closed");
}

void TCPUDGAppBase::socketFailure(int, void *, int code) {
// subclasses may override this function, and add code try to reconnect after a delay.
    EV << "connection broken\n";
    setStatusString("broken");

    numBroken++;
}

void TCPUDGAppBase::finish() {
    std::string modulePath = getFullPath();

    EV << modulePath << ": opened " << numSessions << " sessions\n";
    EV << modulePath << ": sent " << bytesSent << " bytes in " << packetsSent
              << " packets\n";
    EV << modulePath << ": received " << bytesRcvd << " bytes in "
              << packetsRcvd << " packets\n";
}


