//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// //
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//
// Author: Ondrej Vondrous
// Department of Telecommunication Engineering
// Czech Technical University in Prague
// Faculty of Electrical Engineering
// 2015
//
//


#ifndef UDGCONFIG_H_
#define UDGCONFIG_H_

#include <sys/time.h>
#include <queue>
#include <map>
#include <vector>
#include <string>
#include <sys/types.h>

#include "inet/networklayer/common/L3Address.h" //inet 3.x


#define MSG "Local:"<<localAddress <<"  TS:"<< simTime() << "  connId()=="<< socket->getConnectionId() << "  "
#define CONN " "<<socket->getLocalAddress()<<":"<<socket->getLocalPort()<<" --> "<<socket->getRemoteAddress()<<":"<<socket->getRemotePort() << " "
#define RED "\033[0;31m"
#define iRED "\033[1;31m"
#define GREEN "\033[0;32m"
#define YELLOW "\033[0;33m"
#define BLUE "\033[0;34m"
#define iBLUE "\033[1;34m"
#define PURPLE "\033[0;35m"
#define CYAN "\033[0;36m"
#define DEFAULT "\033[0;00m"

/*
 RCol='\e[0m'    # Text Reset

 # Regular           Bold                Underline           High Intensity      BoldHigh Intens     Background          High Intensity Backgrounds
 Bla='\e[0;30m';     BBla='\e[1;30m';    UBla='\e[4;30m';    IBla='\e[0;90m';    BIBla='\e[1;90m';   On_Bla='\e[40m';    On_IBla='\e[0;100m';
 Red='\e[0;31m';     BRed='\e[1;31m';    URed='\e[4;31m';    IRed='\e[0;91m';    BIRed='\e[1;91m';   On_Red='\e[41m';    On_IRed='\e[0;101m';
 Gre='\e[0;32m';     BGre='\e[1;32m';    UGre='\e[4;32m';    IGre='\e[0;92m';    BIGre='\e[1;92m';   On_Gre='\e[42m';    On_IGre='\e[0;102m';
 Yel='\e[0;33m';     BYel='\e[1;33m';    UYel='\e[4;33m';    IYel='\e[0;93m';    BIYel='\e[1;93m';   On_Yel='\e[43m';    On_IYel='\e[0;103m';
 Blu='\e[0;34m';     BBlu='\e[1;34m';    UBlu='\e[4;34m';    IBlu='\e[0;94m';    BIBlu='\e[1;94m';   On_Blu='\e[44m';    On_IBlu='\e[0;104m';
 Pur='\e[0;35m';     BPur='\e[1;35m';    UPur='\e[4;35m';    IPur='\e[0;95m';    BIPur='\e[1;95m';   On_Pur='\e[45m';    On_IPur='\e[0;105m';
 Cya='\e[0;36m';     BCya='\e[1;36m';    UCya='\e[4;36m';    ICya='\e[0;96m';    BICya='\e[1;96m';   On_Cya='\e[46m';    On_ICya='\e[0;106m';
 Whi='\e[0;37m';     BWhi='\e[1;37m';    UWhi='\e[4;37m';    IWhi='\e[0;97m';    BIWhi='\e[1;97m';   On_Whi='\e[47m';    On_IWhi='\e[0;107m';

 */

using namespace std;

enum dir_t {sd, ds};
enum type_t { tcp_t, udp_t };

struct packet_t{
    dir_t dir;
    int id;
    double time;
    u_int64_t size;
    string type;
    int prev;
    int pprev;
};

struct profile_t{
    int id;
    type_t type;                // tcp vs. udp
    inet::L3Address src_ip;
    inet::L3Address dst_ip;
    int src_port;
    int dst_port;
    queue<packet_t *> sd_send_data; //pkts in right order
    queue<packet_t *> ds_send_data; //pkts in right order
    queue<packet_t *> sd_sent_xdata; //pkts in right order
    queue<packet_t *> ds_sent_xdata; //pkts in right order
    map<int,map<int, queue<packet_t *> *> *> reply_data; //random access - [PPrev][Prev][PK Queue]
    double stime; //relative packet timing [simTime()+stime]
};

struct udg_profiles_t{
    float version;
    vector<profile_t *> profiles;
};

class UDGConfig {
public:
    UDGConfig(const cXMLElement *xmlFile, const inet::L3Address ip, const type_t type);
    virtual ~UDGConfig();
    vector<profile_t *> getUDGProfiles(const inet::L3Address addr,const type_t type);
private:
    udg_profiles_t udg_profiles;
    bool readXMLFile(const cXMLElement *xmlFile, const inet::L3Address ip, const type_t type);
};

#endif /* UDGCONFIG_H_ */
